/*@ src/su/gen-cs-ctype.h, generated by su-make-cs-ctype.sh.
 *@ See cs-ctype.c for more */

CTAV(su_CS_CTYPE_NONE == 0);
#undef a_X
#define a_X(X) su_CONCAT(su_CS_CTYPE_,X)
u16 const su__cs_ctype[S8_MAX + 1] = {
a_X(CNTRL),
a_X(CNTRL),
a_X(CNTRL),
a_X(CNTRL),
a_X(CNTRL),
a_X(CNTRL),
a_X(CNTRL),
a_X(CNTRL),
a_X(CNTRL),
a_X(BLANK)|a_X(CNTRL)|a_X(SPACE)|a_X(WHITE),
a_X(CNTRL)|a_X(SPACE)|a_X(WHITE),
a_X(CNTRL)|a_X(SPACE),
a_X(CNTRL)|a_X(SPACE),
a_X(CNTRL)|a_X(SPACE),
a_X(CNTRL),
a_X(CNTRL),
a_X(CNTRL),
a_X(CNTRL),
a_X(CNTRL),
a_X(CNTRL),
a_X(CNTRL),
a_X(CNTRL),
a_X(CNTRL),
a_X(CNTRL),
a_X(CNTRL),
a_X(CNTRL),
a_X(CNTRL),
a_X(CNTRL),
a_X(CNTRL),
a_X(CNTRL),
a_X(CNTRL),
a_X(CNTRL),
a_X(BLANK)|a_X(PRINT)|a_X(SPACE)|a_X(WHITE),
a_X(GRAPH)|a_X(PRINT)|a_X(PUNCT),
a_X(GRAPH)|a_X(PRINT)|a_X(PUNCT),
a_X(GRAPH)|a_X(PRINT)|a_X(PUNCT),
a_X(GRAPH)|a_X(PRINT)|a_X(PUNCT),
a_X(GRAPH)|a_X(PRINT)|a_X(PUNCT),
a_X(GRAPH)|a_X(PRINT)|a_X(PUNCT),
a_X(GRAPH)|a_X(PRINT)|a_X(PUNCT),
a_X(GRAPH)|a_X(PRINT)|a_X(PUNCT),
a_X(GRAPH)|a_X(PRINT)|a_X(PUNCT),
a_X(GRAPH)|a_X(PRINT)|a_X(PUNCT),
a_X(GRAPH)|a_X(PRINT)|a_X(PUNCT),
a_X(GRAPH)|a_X(PRINT)|a_X(PUNCT),
a_X(GRAPH)|a_X(PRINT)|a_X(PUNCT),
a_X(GRAPH)|a_X(PRINT)|a_X(PUNCT),
a_X(GRAPH)|a_X(PRINT)|a_X(PUNCT),
a_X(ALNUM)|a_X(DIGIT)|a_X(GRAPH)|a_X(PRINT)|a_X(XDIGIT),
a_X(ALNUM)|a_X(DIGIT)|a_X(GRAPH)|a_X(PRINT)|a_X(XDIGIT),
a_X(ALNUM)|a_X(DIGIT)|a_X(GRAPH)|a_X(PRINT)|a_X(XDIGIT),
a_X(ALNUM)|a_X(DIGIT)|a_X(GRAPH)|a_X(PRINT)|a_X(XDIGIT),
a_X(ALNUM)|a_X(DIGIT)|a_X(GRAPH)|a_X(PRINT)|a_X(XDIGIT),
a_X(ALNUM)|a_X(DIGIT)|a_X(GRAPH)|a_X(PRINT)|a_X(XDIGIT),
a_X(ALNUM)|a_X(DIGIT)|a_X(GRAPH)|a_X(PRINT)|a_X(XDIGIT),
a_X(ALNUM)|a_X(DIGIT)|a_X(GRAPH)|a_X(PRINT)|a_X(XDIGIT),
a_X(ALNUM)|a_X(DIGIT)|a_X(GRAPH)|a_X(PRINT)|a_X(XDIGIT),
a_X(ALNUM)|a_X(DIGIT)|a_X(GRAPH)|a_X(PRINT)|a_X(XDIGIT),
a_X(GRAPH)|a_X(PRINT)|a_X(PUNCT),
a_X(GRAPH)|a_X(PRINT)|a_X(PUNCT),
a_X(GRAPH)|a_X(PRINT)|a_X(PUNCT),
a_X(GRAPH)|a_X(PRINT)|a_X(PUNCT),
a_X(GRAPH)|a_X(PRINT)|a_X(PUNCT),
a_X(GRAPH)|a_X(PRINT)|a_X(PUNCT),
a_X(GRAPH)|a_X(PRINT)|a_X(PUNCT),
a_X(ALNUM)|a_X(ALPHA)|a_X(GRAPH)|a_X(PRINT)|a_X(UPPER)|a_X(XDIGIT),
a_X(ALNUM)|a_X(ALPHA)|a_X(GRAPH)|a_X(PRINT)|a_X(UPPER)|a_X(XDIGIT),
a_X(ALNUM)|a_X(ALPHA)|a_X(GRAPH)|a_X(PRINT)|a_X(UPPER)|a_X(XDIGIT),
a_X(ALNUM)|a_X(ALPHA)|a_X(GRAPH)|a_X(PRINT)|a_X(UPPER)|a_X(XDIGIT),
a_X(ALNUM)|a_X(ALPHA)|a_X(GRAPH)|a_X(PRINT)|a_X(UPPER)|a_X(XDIGIT),
a_X(ALNUM)|a_X(ALPHA)|a_X(GRAPH)|a_X(PRINT)|a_X(UPPER)|a_X(XDIGIT),
a_X(ALNUM)|a_X(ALPHA)|a_X(GRAPH)|a_X(PRINT)|a_X(UPPER),
a_X(ALNUM)|a_X(ALPHA)|a_X(GRAPH)|a_X(PRINT)|a_X(UPPER),
a_X(ALNUM)|a_X(ALPHA)|a_X(GRAPH)|a_X(PRINT)|a_X(UPPER),
a_X(ALNUM)|a_X(ALPHA)|a_X(GRAPH)|a_X(PRINT)|a_X(UPPER),
a_X(ALNUM)|a_X(ALPHA)|a_X(GRAPH)|a_X(PRINT)|a_X(UPPER),
a_X(ALNUM)|a_X(ALPHA)|a_X(GRAPH)|a_X(PRINT)|a_X(UPPER),
a_X(ALNUM)|a_X(ALPHA)|a_X(GRAPH)|a_X(PRINT)|a_X(UPPER),
a_X(ALNUM)|a_X(ALPHA)|a_X(GRAPH)|a_X(PRINT)|a_X(UPPER),
a_X(ALNUM)|a_X(ALPHA)|a_X(GRAPH)|a_X(PRINT)|a_X(UPPER),
a_X(ALNUM)|a_X(ALPHA)|a_X(GRAPH)|a_X(PRINT)|a_X(UPPER),
a_X(ALNUM)|a_X(ALPHA)|a_X(GRAPH)|a_X(PRINT)|a_X(UPPER),
a_X(ALNUM)|a_X(ALPHA)|a_X(GRAPH)|a_X(PRINT)|a_X(UPPER),
a_X(ALNUM)|a_X(ALPHA)|a_X(GRAPH)|a_X(PRINT)|a_X(UPPER),
a_X(ALNUM)|a_X(ALPHA)|a_X(GRAPH)|a_X(PRINT)|a_X(UPPER),
a_X(ALNUM)|a_X(ALPHA)|a_X(GRAPH)|a_X(PRINT)|a_X(UPPER),
a_X(ALNUM)|a_X(ALPHA)|a_X(GRAPH)|a_X(PRINT)|a_X(UPPER),
a_X(ALNUM)|a_X(ALPHA)|a_X(GRAPH)|a_X(PRINT)|a_X(UPPER),
a_X(ALNUM)|a_X(ALPHA)|a_X(GRAPH)|a_X(PRINT)|a_X(UPPER),
a_X(ALNUM)|a_X(ALPHA)|a_X(GRAPH)|a_X(PRINT)|a_X(UPPER),
a_X(ALNUM)|a_X(ALPHA)|a_X(GRAPH)|a_X(PRINT)|a_X(UPPER),
a_X(GRAPH)|a_X(PRINT)|a_X(PUNCT),
a_X(GRAPH)|a_X(PRINT)|a_X(PUNCT),
a_X(GRAPH)|a_X(PRINT)|a_X(PUNCT),
a_X(GRAPH)|a_X(PRINT)|a_X(PUNCT),
a_X(GRAPH)|a_X(PRINT)|a_X(PUNCT),
a_X(GRAPH)|a_X(PRINT)|a_X(PUNCT),
a_X(ALNUM)|a_X(ALPHA)|a_X(GRAPH)|a_X(LOWER)|a_X(PRINT)|a_X(XDIGIT),
a_X(ALNUM)|a_X(ALPHA)|a_X(GRAPH)|a_X(LOWER)|a_X(PRINT)|a_X(XDIGIT),
a_X(ALNUM)|a_X(ALPHA)|a_X(GRAPH)|a_X(LOWER)|a_X(PRINT)|a_X(XDIGIT),
a_X(ALNUM)|a_X(ALPHA)|a_X(GRAPH)|a_X(LOWER)|a_X(PRINT)|a_X(XDIGIT),
a_X(ALNUM)|a_X(ALPHA)|a_X(GRAPH)|a_X(LOWER)|a_X(PRINT)|a_X(XDIGIT),
a_X(ALNUM)|a_X(ALPHA)|a_X(GRAPH)|a_X(LOWER)|a_X(PRINT)|a_X(XDIGIT),
a_X(ALNUM)|a_X(ALPHA)|a_X(GRAPH)|a_X(LOWER)|a_X(PRINT),
a_X(ALNUM)|a_X(ALPHA)|a_X(GRAPH)|a_X(LOWER)|a_X(PRINT),
a_X(ALNUM)|a_X(ALPHA)|a_X(GRAPH)|a_X(LOWER)|a_X(PRINT),
a_X(ALNUM)|a_X(ALPHA)|a_X(GRAPH)|a_X(LOWER)|a_X(PRINT),
a_X(ALNUM)|a_X(ALPHA)|a_X(GRAPH)|a_X(LOWER)|a_X(PRINT),
a_X(ALNUM)|a_X(ALPHA)|a_X(GRAPH)|a_X(LOWER)|a_X(PRINT),
a_X(ALNUM)|a_X(ALPHA)|a_X(GRAPH)|a_X(LOWER)|a_X(PRINT),
a_X(ALNUM)|a_X(ALPHA)|a_X(GRAPH)|a_X(LOWER)|a_X(PRINT),
a_X(ALNUM)|a_X(ALPHA)|a_X(GRAPH)|a_X(LOWER)|a_X(PRINT),
a_X(ALNUM)|a_X(ALPHA)|a_X(GRAPH)|a_X(LOWER)|a_X(PRINT),
a_X(ALNUM)|a_X(ALPHA)|a_X(GRAPH)|a_X(LOWER)|a_X(PRINT),
a_X(ALNUM)|a_X(ALPHA)|a_X(GRAPH)|a_X(LOWER)|a_X(PRINT),
a_X(ALNUM)|a_X(ALPHA)|a_X(GRAPH)|a_X(LOWER)|a_X(PRINT),
a_X(ALNUM)|a_X(ALPHA)|a_X(GRAPH)|a_X(LOWER)|a_X(PRINT),
a_X(ALNUM)|a_X(ALPHA)|a_X(GRAPH)|a_X(LOWER)|a_X(PRINT),
a_X(ALNUM)|a_X(ALPHA)|a_X(GRAPH)|a_X(LOWER)|a_X(PRINT),
a_X(ALNUM)|a_X(ALPHA)|a_X(GRAPH)|a_X(LOWER)|a_X(PRINT),
a_X(ALNUM)|a_X(ALPHA)|a_X(GRAPH)|a_X(LOWER)|a_X(PRINT),
a_X(ALNUM)|a_X(ALPHA)|a_X(GRAPH)|a_X(LOWER)|a_X(PRINT),
a_X(ALNUM)|a_X(ALPHA)|a_X(GRAPH)|a_X(LOWER)|a_X(PRINT),
a_X(GRAPH)|a_X(PRINT)|a_X(PUNCT),
a_X(GRAPH)|a_X(PRINT)|a_X(PUNCT),
a_X(GRAPH)|a_X(PRINT)|a_X(PUNCT),
a_X(GRAPH)|a_X(PRINT)|a_X(PUNCT),
a_X(CNTRL),
};
#undef a_X
u8 const su__cs_tolower[S8_MAX + 1] = {
'\x00','\x01','\x02','\x03','\x04','\x05','\x06','\x07','\x08','\x09','\x0A',
'\x0B','\x0C','\x0D','\x0E','\x0F','\x10','\x11','\x12','\x13','\x14','\x15',
'\x16','\x17','\x18','\x19','\x1A','\x1B','\x1C','\x1D','\x1E','\x1F','\x20',
'\x21','\x22','\x23','\x24','\x25','\x26','\x27','\x28','\x29','\x2A','\x2B',
'\x2C','\x2D','\x2E','\x2F','\x30','\x31','\x32','\x33','\x34','\x35','\x36',
'\x37','\x38','\x39','\x3A','\x3B','\x3C','\x3D','\x3E','\x3F','\x40','\x61',
'\x62','\x63','\x64','\x65','\x66','\x67','\x68','\x69','\x6A','\x6B','\x6C',
'\x6D','\x6E','\x6F','\x70','\x71','\x72','\x73','\x74','\x75','\x76','\x77',
'\x78','\x79','\x7A','\x5B','\x5C','\x5D','\x5E','\x5F','\x60','\x61','\x62',
'\x63','\x64','\x65','\x66','\x67','\x68','\x69','\x6A','\x6B','\x6C','\x6D',
'\x6E','\x6F','\x70','\x71','\x72','\x73','\x74','\x75','\x76','\x77','\x78',
'\x79','\x7A','\x7B','\x7C','\x7D','\x7E','\x7F',
};
u8 const su__cs_toupper[S8_MAX + 1] = {
'\x00','\x01','\x02','\x03','\x04','\x05','\x06','\x07','\x08','\x09','\x0A',
'\x0B','\x0C','\x0D','\x0E','\x0F','\x10','\x11','\x12','\x13','\x14','\x15',
'\x16','\x17','\x18','\x19','\x1A','\x1B','\x1C','\x1D','\x1E','\x1F','\x20',
'\x21','\x22','\x23','\x24','\x25','\x26','\x27','\x28','\x29','\x2A','\x2B',
'\x2C','\x2D','\x2E','\x2F','\x30','\x31','\x32','\x33','\x34','\x35','\x36',
'\x37','\x38','\x39','\x3A','\x3B','\x3C','\x3D','\x3E','\x3F','\x40','\x41',
'\x42','\x43','\x44','\x45','\x46','\x47','\x48','\x49','\x4A','\x4B','\x4C',
'\x4D','\x4E','\x4F','\x50','\x51','\x52','\x53','\x54','\x55','\x56','\x57',
'\x58','\x59','\x5A','\x5B','\x5C','\x5D','\x5E','\x5F','\x60','\x41','\x42',
'\x43','\x44','\x45','\x46','\x47','\x48','\x49','\x4A','\x4B','\x4C','\x4D',
'\x4E','\x4F','\x50','\x51','\x52','\x53','\x54','\x55','\x56','\x57','\x58',
'\x59','\x5A','\x7B','\x7C','\x7D','\x7E','\x7F',
};
