/*@ gen-tcaps.h, generated by make-tcap-map.pl.
 *@ See termcap.c for more */

static char const a_termcap_namedat[] = {
't','e', 'r','m','c','u','p','\0',
't','i', 's','m','c','u','p','\0',
'k','s', 's','m','k','x','\0',
'k','e', 'r','m','k','x','\0',
'c','e', 'e','l','\0',
'c','h', 'h','p','a','\0',
'c','r', 'c','r','\0',
'l','e', 'c','u','b','1','\0',
'n','d', 'c','u','f','1','\0',
'c','d', 'e','d','\0',
'h','o', 'h','o','m','e','\0',
'c','l', 'c','l','e','a','r','\0',
'a','m', 'a','m','\0',
'Y','E', 's','a','m','\0',
'x','n', 'x','e','n','l','\0',
'C','o', 'c','o','l','o','r','s','\0',
#ifdef mx_HAVE_KEY_BINDINGS
'k','b', 'k','b','s','\0',
'k','D', 'k','d','c','h','1','\0',
'*','4', 'k','D','C','\0',
'k','E', 'k','e','l','\0',
'@','9', 'k','e','x','t','\0',
'k','I', 'k','i','c','h','1','\0',
'#','3', 'k','I','C','\0',
'k','h', 'k','h','o','m','e','\0',
'#','2', 'k','H','O','M','\0',
'@','7', 'k','e','n','d','\0',
'*','7', 'k','E','N','D','\0',
'k','N', 'k','n','p','\0',
'k','P', 'k','p','p','\0',
'k','l', 'k','c','u','b','1','\0',
'#','4', 'k','L','F','T','\0',
'\0','\0', 'k','L','F','T','3','\0',
'\0','\0', 'k','L','F','T','5','\0',
'k','r', 'k','c','u','f','1','\0',
'%','i', 'k','R','I','T','\0',
'\0','\0', 'k','R','I','T','3','\0',
'\0','\0', 'k','R','I','T','5','\0',
'k','d', 'k','c','u','d','1','\0',
'\0','\0', 'k','D','N','\0',
'\0','\0', 'k','D','N','3','\0',
'\0','\0', 'k','D','N','5','\0',
'k','u', 'k','c','u','u','1','\0',
'\0','\0', 'k','U','P','\0',
'\0','\0', 'k','U','P','3','\0',
'\0','\0', 'k','U','P','5','\0',
'k','0', 'k','f','0','\0',
'k','1', 'k','f','1','\0',
'k','2', 'k','f','2','\0',
'k','3', 'k','f','3','\0',
'k','4', 'k','f','4','\0',
'k','5', 'k','f','5','\0',
'k','6', 'k','f','6','\0',
'k','7', 'k','f','7','\0',
'k','8', 'k','f','8','\0',
'k','9', 'k','f','9','\0',
'k',';', 'k','f','1','0','\0',
'F','1', 'k','f','1','1','\0',
'F','2', 'k','f','1','2','\0',
'F','3', 'k','f','1','3','\0',
'F','4', 'k','f','1','4','\0',
'F','5', 'k','f','1','5','\0',
'F','6', 'k','f','1','6','\0',
'F','7', 'k','f','1','7','\0',
'F','8', 'k','f','1','8','\0',
'F','9', 'k','f','1','9','\0',
#endif /* mx_HAVE_KEY_BINDINGS */
};

static struct a_termcap_control const a_termcap_control[] = {
# ifdef mx_HAVE_TERMCAP
{mx_TERMCAP_CAPTYPE_STRING, 0},
{mx_TERMCAP_CAPTYPE_STRING, 8},
{mx_TERMCAP_CAPTYPE_STRING, 16},
{mx_TERMCAP_CAPTYPE_STRING, 23},
# endif
# ifdef mx_HAVE_MLE
{mx_TERMCAP_CAPTYPE_STRING, 30},
{mx_TERMCAP_CAPTYPE_STRING|a_TERMCAP_F_ARG_IDX1, 35},
{mx_TERMCAP_CAPTYPE_STRING, 41},
{mx_TERMCAP_CAPTYPE_STRING|a_TERMCAP_F_ARG_CNT, 46},
{mx_TERMCAP_CAPTYPE_STRING|a_TERMCAP_F_ARG_CNT, 53},
#  ifdef mx_HAVE_TERMCAP
{mx_TERMCAP_CAPTYPE_STRING, 60},
{mx_TERMCAP_CAPTYPE_STRING, 65},
#  endif
{mx_TERMCAP_CAPTYPE_STRING, 72},
# endif
{mx_TERMCAP_CAPTYPE_BOOL|a_TERMCAP_F_QUERY, 80},
{mx_TERMCAP_CAPTYPE_BOOL|a_TERMCAP_F_QUERY, 85},
{mx_TERMCAP_CAPTYPE_BOOL|a_TERMCAP_F_QUERY, 91},
# ifdef mx_HAVE_COLOUR
{mx_TERMCAP_CAPTYPE_NUMERIC|a_TERMCAP_F_QUERY, 98},
# endif
# ifdef mx_HAVE_KEY_BINDINGS
{mx_TERMCAP_CAPTYPE_STRING|a_TERMCAP_F_QUERY, 107},
{mx_TERMCAP_CAPTYPE_STRING|a_TERMCAP_F_QUERY, 113},
{mx_TERMCAP_CAPTYPE_STRING|a_TERMCAP_F_QUERY, 121},
{mx_TERMCAP_CAPTYPE_STRING|a_TERMCAP_F_QUERY, 127},
{mx_TERMCAP_CAPTYPE_STRING|a_TERMCAP_F_QUERY, 133},
{mx_TERMCAP_CAPTYPE_STRING|a_TERMCAP_F_QUERY, 140},
{mx_TERMCAP_CAPTYPE_STRING|a_TERMCAP_F_QUERY, 148},
{mx_TERMCAP_CAPTYPE_STRING|a_TERMCAP_F_QUERY, 154},
{mx_TERMCAP_CAPTYPE_STRING|a_TERMCAP_F_QUERY, 162},
{mx_TERMCAP_CAPTYPE_STRING|a_TERMCAP_F_QUERY, 169},
{mx_TERMCAP_CAPTYPE_STRING|a_TERMCAP_F_QUERY, 176},
{mx_TERMCAP_CAPTYPE_STRING|a_TERMCAP_F_QUERY, 183},
{mx_TERMCAP_CAPTYPE_STRING|a_TERMCAP_F_QUERY, 189},
{mx_TERMCAP_CAPTYPE_STRING|a_TERMCAP_F_QUERY, 195},
{mx_TERMCAP_CAPTYPE_STRING|a_TERMCAP_F_QUERY, 203},
{mx_TERMCAP_CAPTYPE_STRING|a_TERMCAP_F_QUERY, 210},
{mx_TERMCAP_CAPTYPE_STRING|a_TERMCAP_F_QUERY, 218},
{mx_TERMCAP_CAPTYPE_STRING|a_TERMCAP_F_QUERY, 226},
{mx_TERMCAP_CAPTYPE_STRING|a_TERMCAP_F_QUERY, 234},
{mx_TERMCAP_CAPTYPE_STRING|a_TERMCAP_F_QUERY, 241},
{mx_TERMCAP_CAPTYPE_STRING|a_TERMCAP_F_QUERY, 249},
{mx_TERMCAP_CAPTYPE_STRING|a_TERMCAP_F_QUERY, 257},
{mx_TERMCAP_CAPTYPE_STRING|a_TERMCAP_F_QUERY, 265},
{mx_TERMCAP_CAPTYPE_STRING|a_TERMCAP_F_QUERY, 271},
{mx_TERMCAP_CAPTYPE_STRING|a_TERMCAP_F_QUERY, 278},
{mx_TERMCAP_CAPTYPE_STRING|a_TERMCAP_F_QUERY, 285},
{mx_TERMCAP_CAPTYPE_STRING|a_TERMCAP_F_QUERY, 293},
{mx_TERMCAP_CAPTYPE_STRING|a_TERMCAP_F_QUERY, 299},
{mx_TERMCAP_CAPTYPE_STRING|a_TERMCAP_F_QUERY, 306},
{mx_TERMCAP_CAPTYPE_STRING|a_TERMCAP_F_QUERY, 313},
{mx_TERMCAP_CAPTYPE_STRING|a_TERMCAP_F_QUERY, 319},
{mx_TERMCAP_CAPTYPE_STRING|a_TERMCAP_F_QUERY, 325},
{mx_TERMCAP_CAPTYPE_STRING|a_TERMCAP_F_QUERY, 331},
{mx_TERMCAP_CAPTYPE_STRING|a_TERMCAP_F_QUERY, 337},
{mx_TERMCAP_CAPTYPE_STRING|a_TERMCAP_F_QUERY, 343},
{mx_TERMCAP_CAPTYPE_STRING|a_TERMCAP_F_QUERY, 349},
{mx_TERMCAP_CAPTYPE_STRING|a_TERMCAP_F_QUERY, 355},
{mx_TERMCAP_CAPTYPE_STRING|a_TERMCAP_F_QUERY, 361},
{mx_TERMCAP_CAPTYPE_STRING|a_TERMCAP_F_QUERY, 367},
{mx_TERMCAP_CAPTYPE_STRING|a_TERMCAP_F_QUERY, 373},
{mx_TERMCAP_CAPTYPE_STRING|a_TERMCAP_F_QUERY, 380},
{mx_TERMCAP_CAPTYPE_STRING|a_TERMCAP_F_QUERY, 387},
{mx_TERMCAP_CAPTYPE_STRING|a_TERMCAP_F_QUERY, 394},
{mx_TERMCAP_CAPTYPE_STRING|a_TERMCAP_F_QUERY, 401},
{mx_TERMCAP_CAPTYPE_STRING|a_TERMCAP_F_QUERY, 408},
{mx_TERMCAP_CAPTYPE_STRING|a_TERMCAP_F_QUERY, 415},
{mx_TERMCAP_CAPTYPE_STRING|a_TERMCAP_F_QUERY, 422},
{mx_TERMCAP_CAPTYPE_STRING|a_TERMCAP_F_QUERY, 429},
{mx_TERMCAP_CAPTYPE_STRING|a_TERMCAP_F_QUERY, 436},
# endif /* mx_HAVE_KEY_BINDINGS */
};
